function varargout = tspo_ga(varargin)
tic
%% Loading data
load kosmos.txt

departure_ID = kosmos(:,1);
arrival_ID   = kosmos(:,2);
dv           = kosmos(:,3);
dt           = kosmos(:,4);
d_omega      = kosmos(:,5);

total_debris     = 155;
Debris_to_remove = 5;   % Parameter to be changed (4 5 or 8)

time_service    = 20*86400;  % Parameter to be changed (5 10 or 20)
mission_timemax = 300*86400; % Parameter to be changed (180 300 or 365)

g  = 1;

%% Constants and times of references

mu = 398600.4415;
r  = 6378.1363;

v_ref              = sqrt(mu/r);
time_ref           = r/v_ref;
time_servref       = time_service/time_ref;
mission_timemaxref = mission_timemax/time_ref;

%% Configuration of the Algorithm

defaultConfig.popSize     = 1000;
defaultConfig.numIter     = 10;
defaultConfig. showProg   = true;
defaultConfig.showResult  = true;
defaultConfig.showWaitbar = false;

num_redo         = 1; % If there is no reinitialization of population set to 0.
numeroProve      = 2;

% Interpret user configuration inputs
if ~nargin
    userConfig = struct();
elseif isstruct(varargin{1})
    userConfig = varargin{1};
else
    try
        userConfig = struct(varargin{:});
    catch
        error('Expected inputs are either a structure or parameter/value pairs');
    end
end

% Override default configuration with user inputs
configStruct = get_config(defaultConfig,userConfig);

% Extract configuration
popSize     = configStruct.popSize;
numIter     = configStruct.numIter;
showProg    = configStruct.showProg;
showResult  = configStruct.showResult;
showWaitbar = configStruct.showWaitbar;

%% Creation of matrixes & Sanity checks

% Sanity Checks
popSize     = 4*ceil(popSize/4);
numIter     = max(1,round(real(numIter(1))));
showProg    = logical(showProg(1));
showResult  = logical(showResult(1));
showWaitbar = logical(showWaitbar(1));

% Matrixes dmat, tmat, omegamat
dmat     = zeros(total_debris,total_debris);
tmat     = zeros(total_debris,total_debris);
omegamat = zeros(total_debris,total_debris);

row       = 1;
rowt      = 1;
row_omega = 1;

for fileCounter = 0:total_debris:length(departure_ID)-total_debris
    
    dmat(row,:) = dv(fileCounter+1:fileCounter+total_debris);
    row = row + 1;
end

for fileCounter_time = 0:total_debris:length(departure_ID)-total_debris
    
    tmat(rowt,:) = dt(fileCounter_time+1:fileCounter_time+total_debris);
    rowt = rowt + 1;
end

for fileCounter_omega = 0:total_debris:length(departure_ID)-total_debris
    
    omegamat(row_omega,:) = d_omega(fileCounter_omega+1:fileCounter_omega+total_debris);
    row_omega = row_omega + 1;
end

% Creation of Matrixes
resultMatrix   = zeros(numeroProve,4+Debris_to_remove);
totalDist      = zeros(1,popSize);
totalTime      = zeros(1,popSize);
distHistory    = zeros(numeroProve,numIter);
timeHistory    = zeros(numeroProve,numIter);
Result_History = zeros(numIter*numeroProve,2+Debris_to_remove);
newPop         = zeros(popSize,Debris_to_remove);

%% Start with the GA
for prova=1:numeroProve
    prova
    bestIter = inf;
    % Initialize the Population
    pop = zeros(popSize,Debris_to_remove);
    
    for k = 1:popSize
        b = randperm(total_debris);
        c = b(1,1:Debris_to_remove);
        pop(k,:) = c(:);
        
        % Reinitialization of each member of population
        if num_redo>0
            [d_pop_initial] = EvaluatePath_Initial(c,time_servref,mission_timemaxref,tmat,dmat,omegamat,Debris_to_remove,total_debris);
            counter_initial=0;
            while d_pop_initial>=1000 & counter_initial<num_redo
                d_pop_initial_vec(counter_initial+1)=d_pop_initial;
                counter_initial=counter_initial+1;
                b = randperm(total_debris);
                c = b(1,1:Debris_to_remove);
                [d_pop_initial] = EvaluatePath_Initial(c,time_servref,mission_timemaxref,tmat,dmat,omegamat,Debris_to_remove,total_debris);
                d_pop_initial_min=min(d_pop_initial_vec);
                if d_pop_initial<=d_pop_initial_min
                    pop(k,:) = c(:);
                end
            end
        end
    end
    
    %% Start iteration loop & Evaluate each population member
    for iter = 1:numIter
        
        [totalDist,minDist,index,distHistory,globalMin,bestIter,optRoute,totalTime,timeHistory,route_time,Result_History,g] = EvaluatePath(pop,popSize,dmat,tmat,omegamat,numIter,iter,totalDist,distHistory,totalTime,timeHistory,time_servref,mission_timemaxref,Debris_to_remove,prova,total_debris,g,Result_History);
        
        %% Write the results in a matrix
        format long
        resultMatrix(prova,1) = globalMin;
        resultMatrix(prova,2) = route_time;
        resultMatrix(prova,3) = bestIter;
        resultMatrix(prova,4) = index;
        resultMatrix(prova,5:4+Debris_to_remove) = optRoute;
        
        
        %% Bubble Sorting & evaluation of the fitness function
        
        [popSorted4Fitness,choice,roulette_it,totalTime] = BubbleSorting(popSize,pop,totalDist,totalTime,Debris_to_remove);
        
        
        %% APX crossover (PMX)
        
        [child1,child2] = APX_Crossover(choice,popSorted4Fitness,roulette_it,Debris_to_remove,total_debris);
        
        newPop(1:0.3*popSize,:)             = [child1;child2];
        newPop(0.3*popSize+1:0.5*popSize,:) = popSorted4Fitness(0.8*popSize+1:1:popSize,:);
        
        %% Mutation of a part of the best vectors
        
        [child1M,child2M] = Mutation(popSorted4Fitness,Debris_to_remove,total_debris,popSize);
        
        newPop(0.5*popSize+1:0.6*popSize,:) = [child1M;child2M];
        
        %% Single point crossover
        
        [child1SP,child2SP] = SinglePointCrossover(popSorted4Fitness,Debris_to_remove,total_debris,popSize);
        
        newPop(0.6*popSize+1:0.7*popSize,:) = [child1SP;child2SP];
        
        %% Double Point Crossover
        
        [child1DP,child2DP] = DoublePointCrossover(popSorted4Fitness,Debris_to_remove,total_debris,popSize);
        
        newPop(0.7*popSize+1:0.8*popSize,:) = [child1DP;child2DP];
        
        %% Simple inversion of 2 elemnts over the 40 best vectors
        
        invertParent = InvertParent(popSorted4Fitness,Debris_to_remove,popSize);
        
        newPop(0.8*popSize+1:0.9*popSize,:) = invertParent;
        
        %% Introduction of random elements in the new population
        
        for t=0.9*popSize+1:popSize
            newPop(t,:) = randperm(total_debris,Debris_to_remove);
        end
        
        %% Creation of new population
        
        pop = newPop;
        
        %% Elimination of part of population for iter=0.5*numIter
        regCounter = 0;
        
        if iter == 0.5*numIter
            for regCounter=11:popSize
                pop(regCounter,:) = randperm(total_debris,Debris_to_remove);
                
            end
        end
        
        %% Elimination of part of population for iter=0.75*numIter
        
        if iter == 0.75*numIter
            for regCounter=6:popSize
                pop(regCounter,:) = randperm(total_debris,Debris_to_remove);
                
            end
        end
        
    end
end
if showWaitbar
    close(hWait);
end

if showResult
    plot(distHistory.','x','LineWidth',2);
    title('Best Solution History');
end

% Return Output
if nargout
    resultStruct = struct( ...
        'popSize',     popSize, ...
        'numIter',     numIter, ...
        'showProg',    showProg, ...
        'showResult',  showResult, ...
        'showWaitbar', showWaitbar, ...
        'optRoute',    optRoute, ...
        'minDist',     minDist);
    
    varargout = {resultStruct};
end
toc
end

% Subfunction to override the default configuration with user inputs
function config = get_config(defaultConfig,userConfig)

% Initialize the configuration structure as the default
config = defaultConfig;

% Extract the field names of the default configuration structure
defaultFields = fieldnames(defaultConfig);

% Extract the field names of the user configuration structure
userFields = fieldnames(userConfig);
nUserFields = length(userFields);

% Override any default configuration fields with user values
for i = 1:nUserFields
    userField = userFields{i};
    isField = strcmpi(defaultFields,userField);
    if nnz(isField) == 1
        thisField = defaultFields{isField};
        config.(thisField) = userConfig.(userField);
    end
end

end

%% Evaluation of the distance

function [totalDist,minDist,index,distHistory,globalMin,bestIter,optRoute,totalTime,timeHistory,route_time,Result_History,g] = EvaluatePath(pop,popSize,dmat,tmat,omegamat,numIter,iter,totalDist,distHistory,totalTime,timeHistory,time_servref,mission_timemaxref,Debris_to_remove,prova,total_debris,g,Result_History)

globalMin = Inf;
totalDist = zeros(1,popSize);
totalTime = zeros(1,popSize);
for p = 1:popSize
    d = 0; % Open Path
    t = 0;
    for k = 2:Debris_to_remove
        t = t + time_servref;
        t_old = t;
        if length(unique(pop(p,:)))<Debris_to_remove
            pop(p,:) = ChromosomeHealing(pop, p,Debris_to_remove,total_debris);
        end
        t = tmat(pop(p,k-1),pop(p,k));
        omij = abs(omegamat(pop(p,k-1),pop(p,k)));
        if omij == inf
            omij = 1e12;
        end
        if t == inf
            t = 1e12;
            t_old = 1e12;
        end
        d = d + dmat(pop(p,k-1),pop(p,k));
        if t < t_old
            while t < t_old
                t = t + omij;
            end
        end
        if t > mission_timemaxref
            d = 1000+t/107.088;
        end
    end
    
    totalDist(p) = d;
    totalTime(p) = t;
end

% Find the Best Route in the Population
[minDist,index] = min(totalDist);
distHistory(prova,iter) = minDist;
timeHistory(prova,iter) = totalTime(index);
if minDist<1000
    Result_History(g,:) = [minDist totalTime(index) pop(index,:)];
    g=g+1;
end
if minDist < globalMin
    globalMin = minDist;
    bestIter = iter;
    route_time = timeHistory(prova,iter);
    optRoute = pop(index,:);
end

end

%% Evaluation of the distance for the initial population
function [d_pop_initial] = EvaluatePath_Initial(c,time_servref,mission_timemaxref,tmat,dmat,omegamat,Debris_to_remove,total_debris);
d_pop_initial = 0; % Open Path

for k = 2:Debris_to_remove
    t = 0;
    t = t + time_servref;
    t_old = t;
    if length(unique(c(:)))<Debris_to_remove
        c(:) = ChromosomeHealing(c, 1,Debris_to_remove,total_debris);
    end
    t = tmat(c(k-1),c(k));
    omij = abs(omegamat(c(k-1),c(k)));
    if omij == inf
        omij = 1e12;
    end
    
    if t == inf
        t = 1e12;
        t_old = 1e12;
    end
    d_pop_initial = d_pop_initial + dmat(c(k-1),c(k));
    
    if t < t_old
        while t < t_old
            t = t + omij;
        end
    end
    if t > mission_timemaxref
        d_pop_initial = 1000+t/107.088;
    end
end
end

%% Sorting population function

function [popSorted4Fitness,choice,roulette_it,totalTime] = BubbleSorting(popSize,pop,totalDist,totalTime,Debris_to_remove)

popSorted4Fitness = pop;
temp1 = zeros(1,Debris_to_remove);
temp  = 0;
temp2 = 0;
fitness_rank = 1:1:popSize;


for indice=1:popSize
    for indice1=1:popSize-1
        if totalDist(indice1) < totalDist(indice1+1)
            temp = totalDist(indice1);
            temp1 = popSorted4Fitness(indice1,:);
            temp2 = totalTime(indice1);
            totalDist(indice1) = totalDist(indice1+1);
            popSorted4Fitness(indice1,:) = popSorted4Fitness(indice1+1,:);
            totalTime(indice1) = totalTime(indice1+1);
            totalDist(indice1+1) = temp;
            popSorted4Fitness(indice1+1,:) = temp1;
            totalTime(indice1+1) = temp2;
        else if totalDist(indice1) == totalDist(indice1+1) && totalTime(indice1) < totalTime(indice1+1)
                temp = totalDist(indice1);
                temp1 = popSorted4Fitness(indice1,:);
                temp2 = totalTime(indice1);
                totalDist(indice1) = totalDist(indice1+1);
                popSorted4Fitness(indice1,:) = popSorted4Fitness(indice1+1,:);
                totalTime(indice1) = totalTime(indice1+1);
                totalDist(indice1+1) = temp;
                popSorted4Fitness(indice1+1,:) = temp1;
                totalTime(indice1+1) = temp2;
            end
        end
    end
end


roulette_it = 0.3*popSize;
choice = zeros(roulette_it,1);
accumulation = cumsum(fitness_rank);
for i = 1:roulette_it
    p = rand() * accumulation(end);
    chosen_index = -1;
    for index2 = 1 : length(accumulation)
        if (accumulation(index2) > p)
            chosen_index = index2;
            break;
        end
    end
    choice(i) = chosen_index;
end
end

%% Healing of Chromosomes function

function healedChromosome = ChromosomeHealing(chromosome, counter,Debris_to_remove,total_debris)

n = zeros(1,Debris_to_remove);
[C, ia, ic] = unique(chromosome(counter,:),'sorted');
for e=1:length(ia)
    n(ia(e)) = chromosome(counter,ia(e));
end
repetition = chromosome(counter,:)-n;
q = find(repetition);
for u = 1:length(q)
    chromosome(counter,q(u)) = randsample(setdiff(1:total_debris,chromosome(counter,:)), 1);
    
end
healedChromosome = chromosome(counter,:);
end

%% APX Crossover function (PMX)

function [child1,child2] = APX_Crossover(choice,popSorted4Fitness,roulette_it,Debris_to_remove,total_debris)

parent1_index = choice(1:2:roulette_it);
parent2_index = choice(2:2:roulette_it);

parent1Var = popSorted4Fitness(parent1_index,:);
parent2Var = popSorted4Fitness(parent2_index,:);

child1 = zeros(roulette_it/2,Debris_to_remove);
child2 = zeros(roulette_it/2,Debris_to_remove);

for counter=1:length(parent1Var)
    parent1 = parent1Var(counter,:);
    parent2 = parent2Var(counter,:);
    nVar = numel(parent1);
    c  = randsample(nVar,2);
    c1 = min(c);
    c2 = max(c);
    
    L1 = parent1(1:c1);
    L2 = parent2(1:c1);
    
    k1 = parent2(c1+1:c2);
    k2 = parent1(c1+1:c2);
    
    nVar1 = numel(k1);
    nVar2 = numel(k2);
    
    R1 = parent1(c2+1:end);
    R2 = parent2(c2+1:end);
    
    for j=1:nVar
        for i=1:nVar1
            [a,loc1] = ismember(k1(i),L1);
            if a==1
                L1(loc1) = k2(i);
                break;
            end
        end
        for i = 1:nVar1
            [b,loc2] = ismember(k1(i),R1);
            if b==1
                R1(loc2) = k2(i);
            end
        end
    end
    child1(counter,:) = [L1,k1,R1];
    
    for n=1:nVar
        for i=1:nVar2
            [c,loc3] = ismember(k2(i),L2);
            if c==1
                L2(loc3) = k1(i);
            end
        end
        for i=1:nVar2
            [d,loc4] = ismember(k2(i),R2);
            if d==1
                R2(loc4) = k1(i);
            end
        end
    end
    child2(counter,:) = [L2,k2,R2];
    unique(child1(counter,:));
    unique(child2(counter,:));
    
end

% Healing of chromosomes
if length(unique(child1(counter,:)))<Debris_to_remove
    child1(counter,:) = ChromosomeHealing(child1, counter,Debris_to_remove,total_debris);
end

if length(unique(child2(counter,:)))<Debris_to_remove
    child2(counter,:) = ChromosomeHealing(child2, counter,Debris_to_remove,total_debris);
end

end

%% Mutation function

function [child1M,child2M] = Mutation(popSorted4Fitness,Debris_to_remove,total_debris,popSize)

mutationParents = popSorted4Fitness(0.9*popSize+1:popSize,:);
mutationParent1 = mutationParents(1:2:0.1*popSize,:);
mutationParent2 = mutationParents(2:2:0.1*popSize,:);
child1M = zeros(0.05*popSize,Debris_to_remove);
child2M = zeros(0.05*popSize,Debris_to_remove);

for counterM = 1:length(mutationParent1)
    
    child1M(counterM,:) = mutationParent1(counterM,:);
    child2M(counterM,:) = mutationParent2(counterM,:);
    child1M(counterM,1) = randperm(total_debris,1);
    child1M(counterM,2) = randperm(total_debris,1);
    child2M(counterM, Debris_to_remove-1) = randperm(total_debris,1);
    child2M(counterM, Debris_to_remove)   = randperm(total_debris,1);
    
    if length(unique(child1M(counterM,:)))<Debris_to_remove
        child1M(counterM,:) = ChromosomeHealing(child1M, counterM,Debris_to_remove,total_debris);
    end
    
    if length(unique(child2M(counterM,:)))<Debris_to_remove
        child2M(counterM,:) = ChromosomeHealing(child2M, counterM,Debris_to_remove,total_debris);
    end
end
end

%% Single Point Crossover function

function [child1SP,child2SP] = SinglePointCrossover(popSorted4Fitness,Debris_to_remove,total_debris,popSize)

singlePointParents    = popSorted4Fitness(0.9*popSize+1:popSize,:);
singlePointParent1Var = singlePointParents(1:2:0.1*popSize,:);
singlePointParent2Var = singlePointParents(2:2:0.1*popSize,:);
child1SP = zeros(0.05*popSize,Debris_to_remove);
child2SP = zeros(0.05*popSize,Debris_to_remove);

for counterSP=1:length(singlePointParent1Var)
    p1 = singlePointParent1Var(counterSP,:);
    p2 = singlePointParent2Var(counterSP,:);
    
    nVar = numel(p1);
    
    c = randi([1 nVar-1]);
    
    child1SP(counterSP,:) = [p1(1:c) p2(c+1:end)];
    child2SP(counterSP,:) = [p2(1:c) p1(c+1:end)];
    
    if length(unique(child1SP(counterSP,:)))<Debris_to_remove
        child1SP(counterSP,:) = ChromosomeHealing(child1SP, counterSP,Debris_to_remove,total_debris);
    end
    if length(unique(child2SP(counterSP,:)))<Debris_to_remove
        child2SP(counterSP,:) = ChromosomeHealing(child2SP, counterSP,Debris_to_remove,total_debris);
    end
    
end
end

%% Double Point Crossover function

function [child1DP,child2DP] = DoublePointCrossover(popSorted4Fitness,Debris_to_remove,total_debris,popSize)
doublePointParents    = popSorted4Fitness(0.9*popSize+1:popSize,:);
doublePointParent1Var = doublePointParents(1:2:0.1*popSize,:);
doublePointParent2Var = doublePointParents(2:2:0.1*popSize,:);
child1DP = zeros(0.05*popSize,Debris_to_remove);
child2DP = zeros(0.05*popSize,Debris_to_remove);

for counterDP=1:length(doublePointParent1Var)
    p1   = doublePointParent1Var(counterDP,:);
    p2   = doublePointParent2Var(counterDP,:);
    nVar = numel(p1);
    
    c  = randsample(nVar-1,2);
    c1 = min(c);
    c2 = max(c);
    
    child1DP(counterDP,:) = [p1(1:c1) p2(c1+1:c2) p1(c2+1:end)];
    child2DP(counterDP,:) = [p2(1:c1) p1(c1+1:c2) p2(c2+1:end)];
    
    
    if length(unique(child1DP(counterDP,:)))<Debris_to_remove
        child1DP(counterDP,:) = ChromosomeHealing(child1DP, counterDP,Debris_to_remove,total_debris);
    end
    
    if length(unique(child2DP(counterDP,:)))<Debris_to_remove
        child2DP(counterDP,:) = ChromosomeHealing(child2DP, counterDP,Debris_to_remove,total_debris);
    end
    
end
end

%% Inversion function

function invertParent = InvertParent(popSorted4Fitness,Debris_to_remove,popSize)

invertParent = popSorted4Fitness(0.9*popSize+1:popSize,:);
tempInvert = 0;

for counterInvert=1:0.1*popSize
    numbers = randperm(Debris_to_remove,2);
    tempInvert = invertParent(counterInvert,numbers(1,1));
    invertParent(counterInvert,numbers(1,1)) = invertParent(counterInvert,numbers(1,2));
    invertParent(counterInvert,numbers(1,2)) = tempInvert;
end
end